
#include <string.h>
#include <dlfcn.h>

/* for the SDL API */
#include "SDL/SDL.h"
#include "SDL/SDL_audio.h"

/* SDL and config.h define this */
#ifdef HAVE_STDINT_H
#undef HAVE_STDINT_H
#endif

#include "config.h"

/* OpenAL Headers */
#include "alMain.h"
#include "AL/al.h"
#include "AL/alc.h"




static const ALCchar sdl_device[] = "Simple Directmedia Layer"; 
static volatile ALuint load_count;
static void *sdl_handle;
void callback(void *userdata, Uint8 *stream, int len);

#define MAKE_FUNC(x) static typeof(x) * p##x
MAKE_FUNC(SDL_InitSubSystem);
MAKE_FUNC(SDL_OpenAudio);
MAKE_FUNC(SDL_CloseAudio);
MAKE_FUNC(SDL_PauseAudio);
MAKE_FUNC(SDL_QuitSubSystem);
MAKE_FUNC(SDL_WasInit);
MAKE_FUNC(SDL_GetError);
#undef MAKE_FUNC 

static void SDLCALL sdl_callback(void *userdata, Uint8 *stream, int len)
{
	ALCdevice *pDevice = (ALCdevice*)userdata;
	const ALint frameSize = aluChannelsFromFormat(pDevice->Format) *
	aluBytesFromFormat(pDevice->Format);
	aluMixData(pDevice, stream, len/frameSize);
}

void *sdl_load(void)
{
    if(load_count == 0)
    {
        char *str;

#ifdef HAVE_DLFCN_H
        sdl_handle = dlopen("libSDL.so", RTLD_NOW);
        if(!sdl_handle)
            return NULL;
        dlerror();

#define LOAD_FUNC(f) do { \
    p##f = dlsym(sdl_handle, #f); \
    if((str=dlerror()) != NULL) \
    { \
        dlclose(sdl_handle); \
        sdl_handle = NULL; \
        AL_PRINT("Could not load %s from libSDL.so: %s\n", #f, str); \
        return NULL; \
    } \
} while(0)
#else
        str = NULL;
        sdl_handle = (void*)0xDEADBEEF;
	#define LOAD_FUNC(f) p##f = f
#endif

LOAD_FUNC(SDL_InitSubSystem);
LOAD_FUNC(SDL_OpenAudio);
LOAD_FUNC(SDL_CloseAudio);
LOAD_FUNC(SDL_PauseAudio);
LOAD_FUNC(SDL_QuitSubSystem);
LOAD_FUNC(SDL_WasInit);
LOAD_FUNC(SDL_GetError);

#undef LOAD_FUNC
    }
    ++load_count;

    return sdl_handle;
}

static void sdl_unload(void)
{
	if(load_count == 0 || --load_count > 0) return;

	#if defined (HAVE_DLFCN_H)
	dlclose(sdl_handle);
	#endif

	sdl_handle = NULL;
}

static ALCboolean sdl_open_playback(ALCdevice *device, const ALCchar *deviceName){
	sdl_load();
	
	if(!sdl_handle) return ALC_FALSE;
	
	return ALC_TRUE;
}

static void sdl_close_playback(ALCdevice *device){
	pSDL_PauseAudio(1);
	pSDL_CloseAudio();
}

static ALCboolean sdl_reset_playback(ALCdevice *device)
{


    SDL_AudioSpec desired;
    memset(&desired, '\0', sizeof (desired));

    if (pSDL_WasInit(SDL_INIT_AUDIO))
    {
        pSDL_PauseAudio(1);
        pSDL_CloseAudio();
        pSDL_QuitSubSystem(SDL_INIT_AUDIO);
    }

   switch(aluBytesFromFormat(device->Format))
    {
        case 1:
            desired.format = AUDIO_U8;
            break;
        case 4:
            switch(aluChannelsFromFormat(device->Format))
            {
                case 1: device->Format = AL_FORMAT_MONO16; break;
                case 2: device->Format = AL_FORMAT_STEREO16; break;
                case 4: device->Format = AL_FORMAT_QUAD16; break;
                case 6: device->Format = AL_FORMAT_51CHN16; break;
                case 7: device->Format = AL_FORMAT_61CHN16; break;
                case 8: device->Format = AL_FORMAT_71CHN16; break;
            }
            /* fall-through */
        case 2:
            desired.format = AUDIO_S16;
            break;
        default:
            AL_PRINT("Unknown format: 0x%x\n", device->Format);
            return ALC_FALSE;
    }

 desired.freq = device->Frequency;
 desired.channels = aluChannelsFromFormat(device->Format);
 desired.samples = device->UpdateSize * desired.channels;
 desired.callback = sdl_callback;
 desired.userdata = device;

    device->NumUpdates = 2;

    if (pSDL_InitSubSystem(SDL_INIT_AUDIO) == -1)
    {
        AL_PRINT("SDL_InitSubSystem(SDL_INIT_AUDIO) failed: %s\n", pSDL_GetError());
        return ALC_FALSE;
    }

    if (pSDL_OpenAudio(&desired, NULL) == -1)
    {
        AL_PRINT("SDL_OpenAudio failed: %s\n", pSDL_GetError());
        pSDL_QuitSubSystem(SDL_INIT_AUDIO);
        return ALC_FALSE;
    }

    pSDL_PauseAudio(0);
    return ALC_TRUE;
}


static void sdl_stop_playback(ALCdevice *device){
	pSDL_PauseAudio(1);
}

static ALCboolean sdl_open_capture(ALCdevice *pDevice, const ALCchar *deviceName){
	/* No SDL 1.2 implementation of capture devices */
	return ALC_FALSE;
}

static void sdl_close_capture(ALCdevice *pDevice){
	/* No SDL 1.2 implementation of capture devices */
	return;
}
static void sdl_start_capture(ALCdevice *pDevice){
	/* No SDL 1.2 implementation of capture devices */
	return;
}

static void sdl_stop_capture(ALCdevice *pDevice){
	/* No SDL 1.2 implementation of capture devices */
	return;
}

static void sdl_capture_samples(ALCdevice *pDevice, ALCvoid *pBuffer, ALCuint lSamples){
	/* No SDL 1.2 implementation of capture devices */
	return;
}

static ALCuint sdl_available_samples(ALCdevice *pDevice){
	/* No SDL 1.2 implementation of capture devices */
	return 0;
}

BackendFuncs sdl_funcs = {
    sdl_open_playback,
    sdl_close_playback,
    sdl_reset_playback,
    sdl_stop_playback,
    sdl_open_capture,
    sdl_close_capture,
    sdl_start_capture,
    sdl_stop_capture,
    sdl_capture_samples,
    sdl_available_samples
};

void alc_sdl_init(BackendFuncs *func_list){
	*func_list = sdl_funcs;
}

void alc_sdl_deinit(void){
	pSDL_QuitSubSystem(SDL_INIT_AUDIO);
	sdl_unload();
}

void alc_sdl_probe(int type){
	if(type == DEVICE_PROBE)
		AppendDeviceList(sdl_device);
	else if(type == ALL_DEVICE_PROBE)
		AppendAllDeviceList(sdl_device);
}
